import logging

import openpaperwork_core
import openpaperwork_core.promise


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        self.scan_id_to_doc_id = {}
        self.doc_id_to_scan_id = {}

    def get_interfaces(self):
        return ['scan2doc']

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
            {
                'interface': 'page_img',
                'defaults': ['paperwork_backend.model.img'],
            },
            {
                'interface': 'pillow',
                'defaults': ['openpaperwork_core.pillow.img'],
            },
            {
                'interface': 'scan',
                'defaults': ['paperwork_backend.docscan.libinsane'],
            },
            {
                'interface': 'thread',
                'defaults': ['openpaperwork_core.thread.simple'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def scan2doc_scan_id_to_doc_id(self, scan_id):
        try:
            return self.scan_id_to_doc_id[scan_id]
        except KeyError:
            return None

    def scan2doc_doc_id_to_scan_id(self, doc_id):
        try:
            return self.doc_id_to_scan_id[doc_id]
        except KeyError:
            return None

    def scan2doc_promise(self, *args, doc_id=None, doc_url=None, **kwargs):
        """
        The promise returned by this method should be scheduled with
        scan_schedule() to avoid any possible conflict with another
        scan (or scanner lookup).
        """
        if doc_id is not None and doc_url is not None:
            nb_pages = self.core.call_success(
                "doc_get_nb_pages_by_url", doc_url
            )
            new = (nb_pages <= 0) if nb_pages is not None else True
        else:
            (doc_id, doc_url) = self.core.call_success("storage_get_new_doc")
            new = True

        (scan_id, p) = self.core.call_success("scan_promise", *args, **kwargs)

        promise = openpaperwork_core.promise.Promise(
            self.core, self.core.call_all,
            args=("on_scan2doc_start", scan_id, doc_id, doc_url)
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(p)

        self.scan_id_to_doc_id[scan_id] = doc_id
        self.doc_id_to_scan_id[doc_id] = scan_id

        def add_scans_to_doc(args):
            (source, scan_id, imgs) = args
            nb = 0
            for img in imgs:
                nb += 1
                nb_pages = self.core.call_success(
                    "doc_get_nb_pages_by_url", doc_url
                )
                if nb_pages is None:
                    nb_pages = 0
                page_url = self.core.call_success(
                    "page_get_img_url", doc_url, nb_pages, write=True
                )
                self.core.call_success("pillow_to_url", img, page_url)
                self.core.call_all(
                    "mainloop_schedule", self.core.call_all,
                    "on_scan2doc_page_scanned",
                    scan_id, doc_id, doc_url, nb_pages
                )
            return nb

        def drop_scan_id(scan_id, doc_id):
            self.scan_id_to_doc_id.pop(scan_id, None)
            self.doc_id_to_scan_id.pop(doc_id, None)

        def notify_end(scan_id, doc_id):
            self.core.call_all("on_scan2doc_end", scan_id, doc_id, doc_url)
            return (doc_id, doc_url)

        def cancel(exc, scan_id, doc_id):
            drop_scan_id(scan_id, doc_id)
            if new:
                self.core.call_all("storage_delete_doc_id", doc_id)
            raise exc

        def run_transactions(nb_imgs, scan_id, doc_id):
            # start a second promise chain, but scheduled with
            # "transaction_schedule()"
            promise = openpaperwork_core.promise.Promise(
                self.core, drop_scan_id, args=(scan_id, doc_id)
            )
            promise = promise.then(self.core.call_success(
                "transaction_simple_promise",
                [("add" if new else "upd", doc_id),]
            ))
            promise = promise.then(notify_end, scan_id, doc_id)
            promise = promise.catch(cancel, scan_id, doc_id)
            self.core.call_success("transaction_schedule", promise)
            return (doc_id, doc_url)

        promise = promise.then(
            openpaperwork_core.promise.ThreadedPromise(
                self.core, add_scans_to_doc
            )
        )
        promise = promise.then(
            openpaperwork_core.promise.ThreadedPromise(
                self.core, run_transactions, args=(scan_id, doc_id)
            )
        )
        promise = promise.catch(cancel, scan_id, doc_id)
        return promise



"""
This plugin requires pydbus.
If Pydbus is not installed. It just does nothing.
"""

import logging
import os
import socket
import time

try:
    import pydbus
    PYDBUS_AVAILABLE = True
except ImportError:
    PYDBUS_AVAILABLE = False

from .. import PluginBase


LOGGER = logging.getLogger(__name__)


class Plugin(PluginBase):
    PRIORITY = 50

    def get_interfaces(self):
        return ['external_apps']

    def get_deps(self):
        return []

    def _get_file_manager_proxy(self):
        bus = pydbus.SessionBus()
        proxy = bus.get(
            "org.freedesktop.FileManager1",
            "/org/freedesktop/FileManager1"
        )
        iface = proxy['org.freedesktop.FileManager1']
        return iface

    def _open_url(self, func_name, url):
        if not PYDBUS_AVAILABLE:
            LOGGER.info("Pydbus is not available")
            return None
        try:
            uid = "{}{}_TIME{}".format(
                socket.gethostname(), os.getpid(), int(time.time())
            )
            iface = self._get_file_manager_proxy()
            LOGGER.info("Opening %s using Dbus", url)
            getattr(iface, func_name)([url], uid)
        except Exception as exc:
            LOGGER.error(
                "Failed to open %s using Dbus", url, exc_info=exc
            )
            return None
        return True

    def external_app_open_file(self, file_url):
        # doesn't really open the file, but close enough.
        return self._open_url('ShowItems', file_url)

    def external_app_open_folder(self, folder_url):
        return self._open_url('ShowFolders', folder_url)

    def external_app_can_send_as_attachment(self) -> None:
        return None

    def external_app_send_as_attachment(self, uri) -> None:
        return None



import logging

import openpaperwork_core

from ... import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'gtk_settings_scanner_setting',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_settings_scanner',
                'defaults': ['paperwork_gtk.settings.scanner.settings'],
            },
            {
                'interface': 'scan',
                'defaults': ['paperwork_backend.docscan.libinsane'],
            },
        ]

    def complete_scanner_settings(
            self, global_widget_tree, parent_widget_tree,
            list_scanner_promise):
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.settings.scanner",
            "popover.glade"
        )

        widget_tree.get_object("settings_stack").set_visible_child_name(
            "spinner"
        )
        widget_tree.get_object("spinner").start()

        parent_widget_tree.get_object("scanner_device").set_popover(
            widget_tree.get_object("selector")
        )

        list_scanner_promise.then(self._on_scanner_list, widget_tree)

    def _on_scanner_list(self, devs, widget_tree):
        widget_tree.get_object("spinner").stop()
        widget_tree.get_object("settings_stack").set_visible_child_name(
            "selector"
        )
        box = widget_tree.get_object("selector_box")

        radios = []
        # because of the way radio buttons work, we need always at least
        # one choice --> add "no scanner"
        for dev in ([(None, _("No scanner"))] + devs):
            radio = self.core.call_success(
                "gtk_load_widget_tree", "paperwork_gtk.settings.scanner",
                "popover_box.glade"
            )
            radio = radio.get_object("radio")
            radio.set_label(dev[1])
            box.pack_start(radio, expand=False, fill=True, padding=0)
            radios.append((dev[0], radio))

        for (dev_id, radio) in radios[1:]:
            radio.join_group(radios[0][1])

        active = self.core.call_success("config_get", "scanner_dev_id")

        for (dev_id, radio) in radios:
            if active == dev_id:
                radio.set_active(True)
                break

        for (dev_id, radio) in radios:
            radio.connect(
                "toggled", self._on_toggle,
                widget_tree, dev_id, radio.get_label()
            )

    def _on_toggle(
            self, checkbox, widget_tree, dev_id, dev_name):
        LOGGER.info("Selected scanner: %s - %s", dev_id, dev_name)
        widget_tree.get_object("selector").popdown()
        self.core.call_success("config_put", "scanner_dev_id", dev_id)



import array
import io
import logging

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_core.promise


CAIRO_AVAILABLE = False
GLIB_AVAILABLE = False


try:
    import cairo
    CAIRO_AVAILABLE = True
except (ImportError, ValueError):
    pass

try:
    from gi.repository import GObject
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    # dummy so chkdeps can still be called
    class GObject(object):  # type: ignore[no-redef]
        class SignalFlags(object):
            RUN_LAST = 0

        class GObject(object):
            pass


DELAY_SHORT = 0.01
DELAY_LONG = 0.3
LOGGER = logging.getLogger(__name__)

MAX_IMG_DIMENSION = 16 * 1024 - 1


class ImgSurface(object):
    # wrapper so it can be weakref
    def __init__(self, surface):
        self.surface = surface


def limit_img_size(size):
    (width, height) = size
    # Handle Cairo limitation: Dimensions of the image can't exceed 32k
    if width > MAX_IMG_DIMENSION:
        width = MAX_IMG_DIMENSION
    if height > MAX_IMG_DIMENSION:
        height = MAX_IMG_DIMENSION
    return (width, height)


def pillow_to_surface(core, img, intermediate="pixbuf", quality=90):
    """
    Convert a PIL image into a Cairo surface
    """
    if intermediate != "jpeg":
        img.putalpha(256)
        (width, height) = img.size
        img = img.convert("RGBA")
        imgd = img.tobytes('raw', 'BGRA')
        imga = array.array('B', imgd)
        stride = width * 4
        return ImgSurface(cairo.ImageSurface.create_for_data(
            imga, cairo.FORMAT_ARGB32, width, height, stride
        ))

    else:
        img = img.convert("RGB")

        if not hasattr(cairo.ImageSurface, 'set_mime_data'):
            LOGGER.warning(
                "Cairo %s does not support yet 'set_mime_data'."
                " Cannot include image as JPEG in the PDF."
                " Image will be included as PNG (much bigger)",
                cairo.version
            )
            intermediate = 'png'
        else:
            # IMPORTANT: The actual surface will be empty.
            # but mime-data will have attached the correct data
            # to the surface that supports it
            img_surface = ImgSurface(cairo.ImageSurface(
                cairo.FORMAT_RGB24, img.size[0], img.size[1]
            ))
            img_io = io.BytesIO()
            img.save(img_io, format="JPEG", quality=quality)
            img_io.seek(0)
            data = img_io.read()
            img_surface.surface.set_mime_data(cairo.MIME_TYPE_JPEG, data)

    if img_surface is None:
        raise Exception(
            "pillow_to_surface(): unknown intermediate: {}".format(
                intermediate
            )
        )

    core.call_all("on_objref_track", img_surface)
    return img_surface


class CairoRenderer(GObject.GObject):
    __gsignals__ = {
        'getting_size': (GObject.SignalFlags.RUN_LAST, None, ()),
        'size_obtained': (GObject.SignalFlags.RUN_LAST, None, ()),
        'img_obtained': (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    BACKGROUND = (0.5, 0.5, 0.5)
    OUTLINE = (0.5, 0.5, 0.5)

    def __init__(self, core, work_queue_name, file_url):
        super().__init__()
        self.core = core
        self.work_queue_name = work_queue_name
        self.file_url = file_url
        self.size = (0, 0)
        self.zoom = 1.0
        self.cairo_surface = None
        self.visible = False

        promise = openpaperwork_core.promise.Promise(
            self.core, self.emit, args=("getting_size",)
        )
        promise = promise.then(
            core.call_success("url_to_img_size_promise", file_url)
        )
        promise = promise.then(self._set_img_size)
        # Gives back a bit of CPU time to GTK so the GUI remains
        # usable
        promise = promise.then(openpaperwork_core.promise.DelayPromise(
            core, DELAY_SHORT
        ))
        self.get_size_promise = promise
        self.getting_size = False

        # Gives back a bit of CPU time to GTK so the GUI remains
        # usable
        # Give also time so the loading can be cancelled
        promise = openpaperwork_core.promise.DelayPromise(
            core, DELAY_LONG
        )
        promise = promise.then(core.call_success(
            "url_to_cairo_surface_promise", file_url
        ))
        promise = promise.then(self._set_cairo_surface)
        self.render_img_promise = promise
        self.render_job_in_queue = False

    def start(self):
        if self.getting_size:
            # seems render() may be called before start() in some cases
            # --> avoid calling twice work_queue_add_promise() to get the size
            return
        self.core.call_success(
            "work_queue_add_promise",
            self.work_queue_name, self.get_size_promise
        )
        self.getting_size = True

    def render(self, force=False):
        if self.visible and not force:
            return
        self.visible = True
        if self.size == (0, 0):
            self.core.call_all(
                "work_queue_cancel", self.work_queue_name,
                self.get_size_promise
            )
            # re add with a higher priority
            self.core.call_success(
                "work_queue_add_promise",
                self.work_queue_name, self.get_size_promise,
                priority=200
            )
            self.getting_size = True
            return
        if self.render_job_in_queue:
            return
        self.render_job_in_queue = True
        self.core.call_success(
            "work_queue_add_promise",
            self.work_queue_name, self.render_img_promise, priority=100
        )

    def hide(self):
        if not self.visible:
            return
        self.visible = False
        if self.size == (0, 0):
            self.core.call_all(
                "work_queue_cancel", self.work_queue_name,
                self.get_size_promise
            )
            # re add with a lower priority
            self.core.call_success(
                "work_queue_add_promise",
                self.work_queue_name, self.get_size_promise
            )
            self.getting_size = True
        if self.cairo_surface is not None:
            self.cairo_surface.surface.finish()
            self.cairo_surface = None
        self.render_job_in_queue = False
        self.core.call_all(
            "work_queue_cancel", self.work_queue_name, self.render_img_promise
        )

    def close(self):
        self.hide()
        self.core.call_all(
            "work_queue_cancel", self.work_queue_name, self.get_size_promise
        )
        self.get_size_promise = None
        self.getting_size = False
        self.render_img_promise = None

    def _set_img_size(self, size):
        self.size = limit_img_size(size)
        self.getting_size = False
        if self.get_size_promise is None:
            # Document has been closed while we looked for its size
            return
        self.emit("size_obtained")
        if self.visible:
            self.render(force=True)

    def _set_cairo_surface(self, surface):
        self.render_job_in_queue = False
        if not self.visible:  # visibility has changed
            surface.surface.finish()
            return
        self.cairo_surface = surface
        self.emit("img_obtained")

    def _draw(self, cairo_ctx):
        cairo_ctx.save()
        try:
            cairo_ctx.scale(self.zoom, self.zoom)
            cairo_ctx.set_source_surface(self.cairo_surface.surface)
            cairo_ctx.paint()

            size = self.size
            cairo_ctx.set_source_rgb(*self.OUTLINE)
            cairo_ctx.set_line_width(1)
            cairo_ctx.rectangle(
                0, 0,
                size[0] - 1,
                size[1] - 1
            )
            cairo_ctx.stroke()
        finally:
            cairo_ctx.restore()

    def draw(self, cairo_ctx):
        if self.cairo_surface is None:
            cairo_ctx.save()
            try:
                size = self.size
                cairo_ctx.set_source_rgb(*self.BACKGROUND)
                cairo_ctx.rectangle(0, 0, size[0], size[1])
                cairo_ctx.clip()
                cairo_ctx.paint()
            finally:
                cairo_ctx.restore()
        else:
            self._draw(cairo_ctx)

    def blur(self):
        pass

    def unblur(self):
        pass


if GLIB_AVAILABLE:
    GObject.type_register(CairoRenderer)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'chkdeps',
            'cairo_url',
            'pillow_to_surface',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'pillow',
                'defaults': ['openpaperwork_core.pillow.img'],
            },
            {
                'interface': 'work_queue',
                'defaults': ['openpaperwork_core.work_queue.default'],
            },
        ]

    def chkdeps(self, out: dict):
        if not CAIRO_AVAILABLE:
            out['cairo'].update(openpaperwork_core.deps.CAIRO)
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def pillow_to_surface(self, pillow, intermediate="pixbuf", quality=90):
        return pillow_to_surface(self.core, pillow, intermediate, quality)

    def url_to_cairo_surface_promise(self, file_url):
        promise = self.core.call_success("url_to_pillow_promise", file_url)
        promise = promise.then(self.pillow_to_surface)
        return promise

    def url_to_cairo_surface(self, file_url):
        pil_img = self.core.call_success("url_to_pillow", file_url)
        return self.pillow_to_surface(pil_img)

    def cairo_renderer_by_url(self, work_queue_name, file_url, **kwargs):
        return CairoRenderer(self.core, work_queue_name, file_url)



CAIRO = {
    'debian': 'python3-gi-cairo',
    'fedora': 'python3-pycairo',
    'gentoo': 'dev-python/pycairo',  # Python 3 ?
    'linuxmint': 'python-gi-cairo',  # Python 3 ?
    'raspbian': 'python3-gi-cairo',
    'suse': 'python-cairo',  # Python 3 ?
    'ubuntu': 'python3-gi-cairo',
}

GDK = {
    'debian': 'gir1.2-gdkpixbuf-2.0',
    'linuxmint': 'gir1.2-gdkpixbuf-2.0',
    'raspbian': 'gir1.2-gdkpixbuf-2.0',
    'ubuntu': 'gir1.2-gdkpixbuf-2.0',
}

GI = {
    'debian': 'python3-gi',
    'fedora': 'python3-gobject-base',
    'gentoo': 'dev-python/pygobject',  # Python 3 ?
    'linuxmint': 'python3-gi',
    'raspbian': 'python3-gi',
    'suse': 'python-gobject',  # Python 3 ?
    'ubuntu': 'python3-gi',
}

GLIB = {
    'debian': 'gir1.2-glib-2.0',
    'linuxmint': 'gir1.2-glib-2.0',
    'raspbian': 'gir1.2-glib-2.0',
    'ubuntu': 'gir1.2-glib-2.0',
}

PANGO = {
    'debian': 'gir1.2-pango-1.0',
    'linuxmint': 'gir1.2-pango-1.0',
    'raspbian': 'gir1.2-glib-2.0',
    'ubuntu': 'gir1.2-pango-1.0',
}

POPPLER = {
    'debian': 'gir1.2-poppler-0.18',
    'fedora': 'poppler-glib',
    'gentoo': 'app-text/poppler',
    'linuxmint': 'gir1.2-poppler-0.18',
    'raspbian': 'gir1.2-poppler-0.18',
    'suse': 'typelib-1_0-Poppler-0_18',
    'ubuntu': 'gir1.2-poppler-0.18',
}



import collections
import gettext
import importlib
import itertools
import logging
import os
import time


LOGGER = logging.getLogger(__name__)

MINIMUM_CONFIG_PLUGINS = [
    # You also have to provide a plugin providing the interface 'app'
    'openpaperwork_core.archives',
    'openpaperwork_core.cmd.config',
    'openpaperwork_core.cmd.ping',
    'openpaperwork_core.cmd.plugins',
    'openpaperwork_core.config',
    'openpaperwork_core.config.automatic_plugin_reset',
    'openpaperwork_core.config.backend.configparser',
    'openpaperwork_core.data_versioning',
    'openpaperwork_core.frozen',
    'openpaperwork_core.fs.python',
    'openpaperwork_core.logs.archives',
    'openpaperwork_core.mainloop.asyncio',
    'openpaperwork_core.paths.xdg',
    'openpaperwork_core.uncaught_exception',
]


RECOMMENDED_PLUGINS = [
    'openpaperwork_core.external_apps.dbus',
    'openpaperwork_core.external_apps.windows',
    'openpaperwork_core.external_apps.xdg',
    'openpaperwork_core.flatpak',
    'openpaperwork_core.fs.memory',
    'openpaperwork_core.http',
    'openpaperwork_core.i18n.python',
    'openpaperwork_core.l10n.python',
    'openpaperwork_core.perfcheck.log',
    'openpaperwork_core.resources.frozen',
    'openpaperwork_core.resources.setuptools',
    'openpaperwork_core.sqlite',
    'openpaperwork_core.thread.pool',
    'openpaperwork_core.urls',
    'openpaperwork_core.work_queue.default',
]

if os.name != 'nt':
    RECOMMENDED_PLUGINS += [
        'openpaperwork_core.cmd.chkdeps',
    ]


def _(s):
    return gettext.dgettext('openpaperwork_core', s)


class DependencyException(Exception):
    """
    Failed to satisfy dependencies.
    """
    pass


class PluginBase(object):
    """
    Indicates all the methods that must be implemented by any plugin
    managed by OpenPaperwork core. Also provides default implementations
    for each method.
    """

    # Priority defines in which order callbacks will be called.
    # Plugins with higher priorities will have their callbacks called first.
    PRIORITY = 0

    # Convenience for the applications: Indicates if users should be able
    # to enable/disable this plugin in the UI.
    USER_VISIBLE = False

    def __init__(self):
        """
        Called as soon as the module is loaded. Should be as minimal as
        possible. Most of the work should be done in `init()`.
        You *must* *not* rely on any dependencies here.
        """
        self.core = None

    def get_interfaces(self):
        """
        Indicates the list of interfaces implemented by this plugin.
        Interface names are arbitrarily defined. Methods provided by each
        interface are arbitrarily defined (and no checks are done).

        Returns a list of string.
        """
        return []

    def get_deps(self):
        """
        Return the dependencies required by this plugin.

        Example:

        .. code-block:: python

          [
            {
              "interface": "some_interface_name",  # required
              "defaults": ['plugin_a', 'plugin_b'],  # required
              "expected_already_satisfied": False,  # optional, default: True
            },
          ]
        """
        return []

    def init(self, core):
        """
        Plugins can initialize whatever they want here. When called, all
        dependencies have been loaded and initialized, so using them is safe.
        Does nothing by default.
        """
        self.core = core


class Core(object):
    """
    Manage plugins and their callbacks.
    """
    def __init__(self, auto_load_dependencies=False):
        """
        `auto_load_dependencies=True` means that missing dependencies will be
        loaded automatically based on the default plugin list provided by
        plugins. This should be only used for testing.
        """
        self.plugins = {}
        self.initialized = False
        self._to_initialize = set()
        self._initialized = set()  # avoid double-init
        self.interfaces = collections.defaultdict(list)
        self.callbacks = collections.defaultdict(list)
        self.auto_load_dependencies = auto_load_dependencies

        self.log_all = bool(os.getenv("CORE_LOG_ALL", 0))
        self.count_limit_per_second = int(os.getenv("CORE_CALL_LIMIT", 0))
        self.counters_last_reset = 0
        self.counters = collections.defaultdict(lambda: 0)

    def load(self, module_name):
        """
        - Load the specified module
        - Instantiate the class 'Plugin()' of this module
        - Register all the methods of this plugin object (except those starting
          by '_' and those from the class PluginBase) as callbacks

        BEWARE of dependency loops !

        Arguments:
            - module_name: name of the Python module to load
        """
        if module_name in self.plugins:
            return self.plugins[module_name]

        LOGGER.info("Loading plugin '%s' ...", module_name)
        module = importlib.import_module(module_name)
        return self._load_module(module_name, module)

    def _load_module(self, module_name, module):
        """
        should be called from outside for testing only
        """
        if module_name in self.plugins:
            LOGGER.debug("Module %s already loaded", module_name)
            return self.plugins[module_name]

        self.initialized = False

        plugin = module.Plugin()
        self.plugins[module_name] = plugin

        for interface in plugin.get_interfaces():
            LOGGER.debug("- '%s' provides '%s'", str(type(plugin)), interface)
            self.interfaces[interface].append(plugin)

        self._to_initialize.add(plugin)

        LOGGER.info("Plugin '%s' loaded", module_name)
        return plugin

    def _check_deps(self):
        to_examine = [
            (plugin_name, plugin)
            for (plugin_name, plugin) in self.plugins.items()
        ]

        while len(to_examine) > 0:
            (plugin_name, plugin) = to_examine[0]
            to_examine = to_examine[1:]

            LOGGER.info("Examining dependencies of '%s' ...", plugin_name)

            deps = plugin.get_deps()
            for dep in deps:
                interface = dep['interface']
                if len(self.interfaces[interface]) > 0:
                    LOGGER.debug(
                        "- Interface '%s' already provided by %d plugins",
                        interface, len(self.interfaces[interface])
                    )
                    continue

                defaults = dep['defaults']
                if len(defaults) <= 0:
                    continue

                if (not self.auto_load_dependencies
                        and (
                            'expected_already_satisfied' not in dep
                            or dep['expected_already_satisfied']
                        )):
                    LOGGER.warning(
                        "Plugin '{}' requires interface '{}' but no plugins"
                        " provide this interface (suggested: {}). Plugin '{}'"
                        " will not be initialized.".format(
                            plugin_name, interface, defaults, plugin_name
                        )
                    )
                    plugin = self.plugins.pop(plugin_name)
                    self._to_initialize.remove(plugin)
                    # return False to indicate we actually dropped a plugin
                    # and need to reevaluate all the dependencies again.
                    return False
                else:
                    LOGGER.info(
                        "Loading plugins %s to satisfy dependency."
                        " Required by '%s' for interface '%s'",
                        defaults, type(plugin), interface
                    )
                    for default in defaults:
                        to_examine.append((default, self.load(default)))

        return True

    def _register_plugin(self, plugin):
        for attr_name in dir(plugin):
            if attr_name[0] == "_":
                continue
            if attr_name in dir(PluginBase):  # ignore base methods of plugins
                continue
            callback = getattr(plugin, attr_name)
            if not hasattr(callback, '__call__'):
                continue
            LOGGER.debug("- %s.%s()", str(type(plugin)), attr_name)
            self.callbacks[attr_name].append((
                plugin.PRIORITY, str(type(plugin)), callback
            ))
            self.callbacks[attr_name].sort(reverse=True)

    def _init(self, plugin, stack=list()):
        nb = 0
        if plugin in self._initialized:
            return nb

        if plugin in stack:
            LOGGER.error("Dependency loop:")
            for p in itertools.chain(stack, [plugin]):
                LOGGER.error(
                    "- %s %s depends on %s",
                    p, p.get_interfaces(),
                    [d['interface'] for d in p.get_deps()]
                )
            raise DependencyException("Dependency loop: %s" % str(stack))

        stack.append(plugin)

        self.initialized = True

        deps = plugin.get_deps()
        for dep in deps:
            dep_plugins = self.interfaces[dep['interface']]
            for dep_plugin in dep_plugins:
                nb += self._init(dep_plugin, stack)

        LOGGER.info("Initializing plugin '%s' ...", type(plugin))
        stack.remove(plugin)
        plugin.init(self)

        self._register_plugin(plugin)
        nb += 1

        self._initialized.add(plugin)
        return nb

    def init(self):
        """
        - Make sure all the dependencies of all the plugins are satisfied.
        - Call the method init() of each plugin following the dependency
          order (those without dependencies are called first).

        BEWARE of dependency loops !
        """
        LOGGER.info("Initializing all plugins")
        while not self._check_deps():
            pass
        nb = 0
        for plugin in self._to_initialize:
            nb += self._init(plugin)
        self._to_initialize = set()
        LOGGER.info("%d plugins initialized", nb)

    def get_by_name(self, module_name):
        """
        Returns a Plugin instance based on the corresponding module name
        (assuming it has been loaded).

        You shouldn't use this function, except for:
        - unit tests
        - configuration (see cmd.plugins)
        """
        return self.plugins[module_name]

    def get_by_interface(self, interface_name):
        return self.interfaces[interface_name]

    def get_plugins(self):
        """
        You shouldn't use this function, except for:
        - unit tests
        - configuration (see cmd.plugins)
        """
        return dict(self.plugins)

    def _check_call_limit(self, callback_name):
        if self.count_limit_per_second <= 0:
            return
        now = time.time()
        if now - self.counters_last_reset >= 1.0:
            self.counters = collections.defaultdict(lambda: 0)
            self.counters_last_reset = now
        self.counters[callback_name] += 1
        if self.counters[callback_name] >= self.count_limit_per_second:
            raise Exception(
                "Too many calls to '{}' (>= {}) in one second".format(
                    callback_name, self.count_limit_per_second
                )
            )

    def call_all(self, callback_name, *args, **kwargs):
        """
        Call all the methods of all the plugins that have `callback_name`
        as name. Arguments are passed as is. Returned values are dropped
        (use callbacks for return values if required)

        Method call order is defined by the plugin priorities: Plugins with
        a higher priority get their methods called first.

        When we need a return value from callbacks called with `call_all()`,
        we need a way to get the results from all of them. The usual way to do
        that is to instantiate an empty `list` or `set`, and pass it as first
        argument of the callbacks (argument `out`). Callbacks can then
        complete this list or set using `list.append()` or `set.add()`.

        .. uml::

           Caller -> Core: call "func"
           Core -> "Plugin A": plugin.func()
           Core <- "Plugin A": returns "something_a"
           Core -> "Plugin B": plugin.func()
           Core <- "Plugin B": returns "something_b"
           Core -> "Plugin C": plugin.func()
           Core <- "Plugin C": returns "something_c"
           Caller <- Core: returns 3
        """
        if self.log_all:
            print(
                "[{}] call_all({}, args={}, kwargs={})".format(
                    time.time(), callback_name, args, kwargs
                )
            )

        assert \
            self.initialized, \
            "A plugin has been loaded without being initialized." \
            " Call core.init() first"

        self._check_call_limit(callback_name)

        callbacks = self.callbacks[callback_name]
        if len(callbacks) <= 0:
            if callback_name.startswith("on_"):
                # those are 'observer' callback. If nobody is observing,
                # it's usually fine.
                log_method = LOGGER.debug
            else:
                log_method = LOGGER.warning
            log_method("No method '%s' found", callback_name)
            return 0
        for (priority, plugin, callback) in callbacks:
            if self.log_all:
                print(
                    "[{}] call_all({}, args={}, kwargs={}) -> {}:{}".format(
                        time.time(), callback_name, args, kwargs,
                        priority, callback
                    )
                )
            callback(*args, **kwargs)
        return len(callbacks)

    def call_one(self, callback_name, *args, **kwargs):
        """
        Look for a plugin method called `callback_name` and calls it.
        Raises an error if no such method exists. If many exists,
        call one at random. Returns the value return by the callback.

        Method call order is defined by the plugin priorities: Plugins with
        a higher priority get their methods called first.

        .. uml::

           Caller -> Core: call "func"
           Core -> "Plugin A": plugin.func()
           Core <- "Plugin A": returns X
           Caller <- Core: returns X

        You're advised to use `call_all()` or `call_success()` instead
        whenever possible. This method is only provided as convenience for
        when you're fairly sure there should be only one plugin with such
        callback (example: mainloop plugins).
        """
        assert \
            self.initialized, \
            "A plugin has been loaded without being initialized." \
            " Call core.init() first"

        self._check_call_limit(callback_name)
        if self.log_all:
            print(
                "[{}] call_one({}, args={}, kwargs={})".format(
                    time.time(), callback_name, args, kwargs
                )
            )
        callbacks = self.callbacks[callback_name]
        if len(callbacks) <= 0:
            raise IndexError(
                "No method '{}' found !".format(callback_name)
            )
        if self.log_all:
            print(
                "[{}] call_one({}, args={}, kwargs={}) -> {}:{}".format(
                    time.time(), callback_name, args, kwargs,
                    callbacks[0][0], callbacks[0][2]
                )
            )
        return callbacks[0][2](*args, **kwargs)

    def call_success(self, callback_name, *args, **kwargs):
        """
        Call methods of all the plugins that have `callback_name`
        as name until one of them return a value that is not None.
        Arguments are passed as is. First value to be different
        from None is returned. If none of the callbacks returned
        a value different from None or if no callback has the
        specified name, this method will return None.

        Method call order is defined by the plugin priorities: Plugins with
        a higher priority get their methods called first.

        Callbacks should never raise any exception.

        .. uml::

           Caller -> Core: call "func"
           Core -> "Plugin A": plugin.func()
           Core <- "Plugin A": returns None
           Core -> "Plugin B": plugin.func()
           Core <- "Plugin B": returns None
           Core -> "Plugin C": plugin.func()
           Core <- "Plugin C": returns "something"
           Caller <- Core: returns "something"
        """
        assert \
            self.initialized, \
            "A plugin has been loaded without being initialized." \
            " Call core.init() first"

        self._check_call_limit(callback_name)
        if self.log_all:
            print(
                "[{}] call_one({}, args={}, kwargs={})".format(
                    time.time(), callback_name, args, kwargs
                )
            )
        callbacks = self.callbacks[callback_name]
        if len(callbacks) <= 0:
            LOGGER.warning("No method '%s' found", callback_name)
        for (priority, plugin, callback) in callbacks:
            if self.log_all:
                msg = "[{}] call_success({}, args={}, kwargs={}) -> {}:{}"
                print(msg.format(
                    time.time(), callback_name, args, kwargs, priority,
                    callback
                ))
            r = callback(*args, **kwargs)
            if r is not None:
                return r
        return None

    def get_deps(self, plugin_name):
        plugin = self.plugins[plugin_name]
        for dep in plugin.get_deps():
            if 'interface' not in dep:
                raise KeyError(
                    "Missing interface in dependency list of plugin"
                    " '{}'".format(plugin_name)
                )
            if 'defaults' not in dep:
                raise KeyError(
                    "Missing default plugins in dependency list of plugin"
                    " '{}' (interface={})".format(
                        plugin_name, dep['interface']
                    )
                )

            yield {
                'interface': dep['interface'],
                'actives': {
                    x.__module__
                    for x in self.get_by_interface(dep['interface'])
                },
                'defaults': set(dep['defaults']),
            }

    def get_active_plugins(self):
        return self.plugins.keys()



import itertools
import json
import logging
import threading

try:
    import gi
    gi.require_version('Libinsane', '1.0')
    from gi.repository import GObject
    GI_AVAILABLE = True
except (ImportError, ValueError):
    GI_AVAILABLE = False

    # dummy so chkdeps can still be called
    class GObject(object):
        class GObject(object):
            pass

try:
    from gi.repository import Libinsane
    LIBINSANE_AVAILABLE = True
except (ImportError, ValueError):
    LIBINSANE_AVAILABLE = False

    # dummy so chkdeps can still be called
    class Libinsane(object):
        class LogLevel(object):
            DEBUG = 0
            INFO = 1
            WARNING = 2
            ERROR = 3

        class Logger(object):
            pass


import PIL
import PIL.Image

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_core.promise

from .. import _


LOGGER = logging.getLogger(__name__)
SCAN_ID_GENERATOR = itertools.count()

# Prevent closing or other operations on scanner and source instances when
# they are being used (scanning)
LOCK = threading.RLock()


class LibinsaneLogger(GObject.GObject, Libinsane.Logger):
    LEVELS = {
        Libinsane.LogLevel.DEBUG: LOGGER.debug,
        Libinsane.LogLevel.INFO: LOGGER.info,
        Libinsane.LogLevel.WARNING: LOGGER.warning,
        Libinsane.LogLevel.ERROR: LOGGER.error,
    }

    min_level = Libinsane.LogLevel.DEBUG

    def do_log(self, lvl, msg):
        if lvl < self.min_level:
            return
        self.LEVELS[lvl]("[LibInsane] " + msg)


def raw_to_img(params, img_bytes):
    fmt = params.get_format()
    assert fmt == Libinsane.ImgFormat.RAW_RGB_24
    w = params.get_width()
    h = int(len(img_bytes) / 3 / w)
    LOGGER.info(
        "raw_to_img(): w=%s, h=%d, len(img_bytes)=%d",
        w, h, len(img_bytes)
    )
    mode = "RGB"
    return PIL.Image.frombuffer(mode, (w, h), img_bytes, "raw", mode, 0, 1)


class ImageAssembler(object):
    MIN_CHUNK_SIZE = 64 * 1024

    def __init__(self, line_width):
        # 'Pieces' are pieces of the images that may or may not contain
        # full lines of pixels (or even partial pixel)

        # 'chunk': We want to provide the GUI with a preview of the scan.
        # To keep things simple, we provide chunk of the image and those
        # chunk must contain only entire lines of pixels.

        # and then we must also provide the whole image at the end.

        self.w = line_width  # in bytes

        self.current_piece = []
        self.all_chunks = []

    def add_piece(self, piece):
        self.current_piece.append(piece)
        lcurrent_piece = sum((len(c) for c in self.current_piece))
        if lcurrent_piece < self.w or lcurrent_piece < self.MIN_CHUNK_SIZE:
            return

        current_piece = b"".join(self.current_piece)
        chunkable = lcurrent_piece - (lcurrent_piece % self.w)
        new_chunk = current_piece[:chunkable]
        new_piece = current_piece[chunkable:]

        self.all_chunks.append(new_chunk)
        if len(new_piece) <= 0:
            self.current_piece = []
        else:
            self.current_piece = [new_piece]

    def get_last_chunk(self):
        if len(self.all_chunks) <= 0:
            return None
        return self.all_chunks[-1]

    def get_image(self):
        return b"".join(self.all_chunks) + b"".join(self.current_piece)


class Source(object):
    def __init__(self, core, scanner, source):
        self.core = core
        self.scanner = scanner
        self.source_id = source.get_name()
        self.source = source

    def __str__(self):
        return "{}:{}".format(str(self.scanner), self.source_id)

    def set_as_default(self):
        self.core.call_all(
            "config_put", "scanner_source_id", self.source_id
        )

    def _get_option(self, name):
        LOGGER.info(
            "Looking for possible values for option '%s' on %s"
            " : %s ...", name, str(self.scanner), self.source_id
        )
        options = self.source.get_options()
        options = {opt.get_name(): opt for opt in options}
        return options[name]

    def _get_opt_constraint(self, name):
        with LOCK:
            opt = self._get_option(name)
            constraint = opt.get_constraint()
            LOGGER.info(
                "%s : %s : %s : Possible values: %s",
                str(self.scanner), self.source_id, name, constraint
            )
            return constraint

    def get_modes(self):
        return self._get_opt_constraint('mode')

    def get_resolutions(self):
        return self._get_opt_constraint('resolution')

    def get_resolutions_promise(self):
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self.get_resolutions
        )

    def set_default_resolution(self, resolution):
        self.core.call_all(
            "config_put", "scanner_resolution", int(resolution)
        )

    def scan(
                self, scan_id=None, resolution=None, max_pages=9999,
                close_on_end=False
            ):
        """
        Returns the source, the scan ID and an image generator
        """
        with LOCK:
            if scan_id is None:
                scan_id = next(SCAN_ID_GENERATOR)

            LOGGER.info("(id=%s) Setting scan options ...", scan_id)
            if resolution is None:
                resolution = self.core.call_success(
                    "config_get", "scanner_resolution"
                )
            mode = self.core.call_success("config_get", "scanner_mode")

            options = self.source.get_options()

            opts = {opt.get_name(): opt for opt in options}
            if 'resolution' in opts:
                opts['resolution'].set_value(resolution)
            if 'mode' in opts:
                try:
                    opts['mode'].set_value(mode)
                except Exception as exc:
                    LOGGER.warning(
                        "Failed to set scan mode", exc_info=exc
                    )
                    # will try to scan anyway

            extrasettings = self.core.call_success(
                "config_get", "scanner_extra_settings"
            )

            for (setting, value) in extrasettings.items():
                if setting in opts:
                    try:
                        opts[setting].set_value(value)
                    except Exception as exc:
                        LOGGER.warning(
                            "Failed to set scan setting %s=%s",
                            setting, value, exc_info=exc
                        )
                else:
                    LOGGER.warning("No such scanner setting: " + setting)

            imgs = self._scan(scan_id, resolution, max_pages, close_on_end)
            return (self, scan_id, imgs)

    def _scan(self, scan_id, resolution, max_pages, close_on_end=False):
        """
        Returns an image generator
        """
        # keep in mind that we are in a thread here, but listeners
        # must be called from the main loop
        LOGGER.info(
            "(id=%s) Scanning at resolution %d dpi ...",
            scan_id, resolution
        )

        has_started = False
        session = None
        try:
            page_nb = 0

            self.core.call_one(
                "mainloop_execute", self.core.call_all,
                "on_scan_feed_start", scan_id
            )
            self.core.call_success(
                "mainloop_schedule", self.core.call_all,
                "on_progress", "scan", 0.0, _("Starting scan ...")
            )
            session = self.source.scan_start()

            while not session.end_of_feed() and page_nb < max_pages:
                self.core.call_success(
                    "mainloop_schedule", self.core.call_all,
                    "on_progress", "scan", 0.0,
                    _("Scanning page %d ...") % (page_nb + 1)
                )

                scan_params = session.get_scan_parameters()
                LOGGER.info(
                    "Expected scan parameters: %s ; %dx%d = %d bytes",
                    scan_params.get_format(),
                    scan_params.get_width(), scan_params.get_height(),
                    scan_params.get_image_size()
                )
                self.core.call_success(
                    "mainloop_schedule", self.core.call_all,
                    "on_scan_page_start", scan_id, page_nb, scan_params
                )

                assert (
                    scan_params.get_format()
                    == Libinsane.ImgFormat.RAW_RGB_24
                )
                image = ImageAssembler(scan_params.get_width() * 3)
                last_chunk = None
                nb_lines = 0
                total_lines = scan_params.get_height()
                buffer_size = (scan_params.get_width() * 3) + 1

                LOGGER.info("Scanning page %d/%d ...", page_nb, max_pages)
                while not session.end_of_page():
                    new_piece = session.read_bytes(buffer_size).get_data()

                    if not has_started:
                        # Mark the application as busy until we get the first
                        # read(). This is the only reliable time to be
                        # sure scanning is actually started.
                        self.core.call_success(
                            "mainloop_schedule", self.core.call_all,
                            "on_scan_started", scan_id
                        )
                        has_started = True

                    image.add_piece(new_piece)

                    chunk = image.get_last_chunk()
                    if chunk is not last_chunk:
                        last_chunk = chunk
                        pil = raw_to_img(scan_params, chunk)
                        nb_lines += pil.size[1]
                        progress = nb_lines / total_lines
                        if progress >= 1.0:
                            progress = 0.999
                        self.core.call_success(
                            "mainloop_schedule", self.core.call_all,
                            "on_progress", "scan", progress,
                            _("Scanning page %d ...") % (page_nb + 1)
                        )
                        self.core.call_success(
                            "mainloop_schedule", self.core.call_all,
                            "on_scan_chunk", scan_id, scan_params, pil
                        )

                LOGGER.info("Page %d/%d scanned", page_nb, max_pages)
                self.core.call_success(
                    "mainloop_schedule", self.core.call_all,
                    "on_progress", "scan", 0.999,
                    _("Scanning page %d ...") % (page_nb + 1)
                )
                img = raw_to_img(scan_params, image.get_image())
                yield img
                self.core.call_success(
                    "mainloop_schedule", self.core.call_all,
                    "on_scan_page_end", scan_id, page_nb, img
                )
                page_nb += 1
            LOGGER.info("End of feed")

            self.core.call_success(
                "mainloop_schedule", self.core.call_all,
                "on_scan_feed_end", scan_id
            )
        finally:
            self.core.call_success(
                "mainloop_schedule", self.core.call_all,
                "on_progress", "scan", 1.0
            )
            if session is not None:
                session.cancel()
            if close_on_end:
                self.close()

    def scan_promise(self, *args, scan_id=None, **kwargs):
        if scan_id is None:
            scan_id = next(SCAN_ID_GENERATOR)
        kwargs['scan_id'] = scan_id
        return (
            scan_id,
            openpaperwork_core.promise.ThreadedPromise(
                self.core, self.scan, args=args, kwargs=kwargs
            )
        )

    def close(self, *args, **kwargs):
        with LOCK:
            self.scanner.close()
            # return the args for convience when used with promises
            if len(args) == 1 and len(kwargs) == 0:
                return args
            return (args, kwargs)


class Scanner(object):
    def __init__(self, core, scanner):
        self.core = core
        self.dev_id = scanner.get_name()
        self.dev = scanner
        self.sources = None  # WORKAROUND(Jflesch): just to keep ref on them

    def __str__(self):
        return self.dev_id

    def __del__(self):
        if self.dev is not None:
            # Shouldn't happen.
            LOGGER.warning(
                "Scanner(%s, %s) is being garbage-collected",
                self.dev_id, id(self)
            )
        self.close()

    def close(self, *args, **kwargs):
        with LOCK:
            if self.dev is not None:
                LOGGER.info("Closing device %s (%s)", self.dev_id, id(self))
                self.dev.close()
                self.dev = None
            # return the args for convenience when used with promises
            if len(args) == 1 and len(kwargs) == 0:
                return args
            return (args, kwargs)

    def get_sources(self):
        with LOCK:
            LOGGER.info("Looking for scan sources on %s ...", self.dev_id)
            sources = self.dev.get_children()
            sources = [
                Source(self.core, self, source)
                for source in sources
            ]
            self.sources = {
                source.source_id: source
                for source in sources
            }
            LOGGER.info("%d sources found: %s", len(sources), sources)
            return self.sources

    def get_sources_promise(self):
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self.get_sources
        )

    def get_source(self, source_id):
        sources = self.get_sources()
        src = sources[source_id]
        return src

    def set_as_default(self):
        self.core.call_all(
            "config_put", "scanner_dev_id", 'libinsane:' + self.dev_id
        )


class BugReportCollector(object):
    def __init__(self, plugin, update_args):
        self.core = plugin.core
        self.plugin = plugin
        self.update_args = update_args

    def _notify(self, msg):
        self.core.call_success(
            "mainloop_schedule", self.core.call_all,
            "bug_report_update_attachment", "scanner",
            {"file_url": msg},
            *self.update_args
        )

    @staticmethod
    def _get_error_proof(func):
        try:
            r = func()
            if type(r) is bool or type(r) is int or type(r) is float:
                return r
            return str(r)
        except Exception as exc:
            return str(exc)

    def _collect_opt_info(self, opt, out: dict):
        out['name'] = self._get_error_proof(opt.get_name)
        out['title'] = self._get_error_proof(opt.get_title)
        out['description'] = self._get_error_proof(opt.get_desc)
        out['capabilities'] = self._get_error_proof(opt.get_capabilities)
        out['unit'] = self._get_error_proof(opt.get_value_unit)
        out['constraint_type'] = self._get_error_proof(opt.get_constraint_type)
        out['constraint'] = self._get_error_proof(opt.get_constraint)
        out['is_readable'] = self._get_error_proof(opt.is_readable)
        out['is_writable'] = self._get_error_proof(opt.is_writable)
        out['value'] = self._get_error_proof(opt.get_value)

    def _collect_item_info(self, item, base_name, out: dict):
        if item is None or item is False:
            out['reachable'] = False
            return

        out['reachable'] = True

        try:
            name = item.get_name()
            if base_name is not None and base_name != "":
                name = base_name + "/" + name
            self._notify(_("Examining %s") % name)
            out['name'] = item.get_name()
            out['options'] = {}
            out['children'] = {}

            options = item.get_options()
            for opt in options:
                out['options'][opt.get_name()] = {}
                self._collect_opt_info(opt, out['options'][opt.get_name()])

            children = item.get_children()
            for child in children:
                out['children'][child.get_name()] = {}
                self._collect_item_info(
                    child, name, out['children'][child.get_name()]
                )
        finally:
            item.close()

    def _write_scanners_info_to_tmp_file(self, infos):
        infos = json.dumps(
            infos, indent=4, separators=(",", ": "), sort_keys=True
        )
        (file_url, fd) = self.core.call_success(
            "fs_mktemp", prefix="scanner_", suffix=".json", mode="w",
            on_disk=True
        )
        with fd:
            fd.write(infos)
        return file_url

    @staticmethod
    def _collect_get_device(libinsane, dev_id):
        try:
            return libinsane.get_device(dev_id)
        except Exception as exc:
            LOGGER.error("Failed to get device [%s]", dev_id, exc_info=exc)
            return False  # can't return None or it will be ignored

    def _collect_all_info(self, scanners):
        out = {}
        promise = openpaperwork_core.promise.Promise(self.core)
        for dev in scanners:
            dev_id = dev[0]
            dev_name = dev[1]
            out[dev_id] = {
                'listing_name': dev_name
            }
            promise = promise.then(
                openpaperwork_core.promise.ThreadedPromise(
                    self.core, self._collect_get_device,
                    args=(self.plugin.libinsane, dev_id[len("libinsane:"):],)
                )
            )
            promise = promise.then(openpaperwork_core.promise.ThreadedPromise(
                self.core, self._collect_item_info, args=("", out[dev_id],)
            ))
            promise = promise.then(lambda *args, **kwargs: None)

        promise = promise.then(
            openpaperwork_core.promise.ThreadedPromise(
                self.core, self._write_scanners_info_to_tmp_file, args=(out,)
            )
        )
        promise = promise.then(
            lambda file_url: self.core.call_all(
                "bug_report_update_attachment", "scanner", {
                    'file_url': file_url,
                    'file_size': self.core.call_success(
                        'fs_getsize', file_url
                    ),
                }, *self.update_args
            )
        )
        self.core.call_success("scan_schedule", promise)

    def run(self):
        promise = openpaperwork_core.promise.Promise(
            self.core, self.core.call_all,
            args=(
                "bug_report_update_attachment", "scanner",
                {"file_url": _("Getting scanner list ...")},
                *self.update_args
            )
        )
        promise = promise.then(self.plugin.scan_list_scanners_promise())
        promise = promise.then(openpaperwork_core.promise.ThreadedPromise(
            self.core, self._collect_all_info
        ))
        self.core.call_success("scan_schedule", promise)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()

        # Looking for devices twice on Linux tends to crash ...
        self.devices_cache = []

        LOGGER.info("Initializing Libinsane ...")
        self.libinsane_logger = LibinsaneLogger()
        self.libinsane = None
        if LIBINSANE_AVAILABLE:
            Libinsane.register_logger(self.libinsane_logger)
            self.libinsane = Libinsane.Api.new_safebet()
            LOGGER.info(
                "Libinsane %s initialized", self.libinsane.get_version()
            )

        self._last_scanner = None

    def get_interfaces(self):
        return [
            "bug_report_attachments",
            "chkdeps",
            "scan",
            "stats",
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'work_queue',
                'defaults': ['openpaperwork_core.work_queue.default'],
            },
        ]

    def init(self, core):
        super().init(core)

        self.core.call_all("work_queue_create", "scanner")

        settings = {
            'scanner_dev_id': self.core.call_success(
                "config_build_simple", "scanner",
                "dev_id", lambda: None
            ),
            'scanner_source_id': self.core.call_success(
                "config_build_simple", "scanner",
                "source", lambda: None
            ),
            'scanner_resolution': self.core.call_success(
                "config_build_simple", "scanner",
                "resolution", lambda: 300
            ),
            'scanner_mode': self.core.call_success(
                "config_build_simple", "scanner",
                "mode", lambda: "Color"
            ),
            'scanner_extra_settings': self.core.call_success(
                "config_build_simple", "scanner",
                "extrasettings", lambda: {}
            ),

        }
        for (k, setting) in settings.items():
            self.core.call_all(
                "config_register", k, setting
            )

    def chkdeps(self, out: dict):
        if not GI_AVAILABLE:
            out['gi'].update(openpaperwork_core.deps.GI)
        if not LIBINSANE_AVAILABLE:
            out['libinsane'] = {
                'debian': 'gir1.2-libinsane-1.0',
                'linuxmint': 'gir1.2-libinsane-1.0',
                'raspbian': 'gir1.2-libinsane-1.0',
                'ubuntu': 'gir1.2-libinsane-1.0',
            }

    def scan_schedule(self, promise):
        """
        Any promise or chain of promises related to scanners must *always*
        be run sequentially to avoid crashes. Otherwise, 2 threaded promises
        could run in parrallel. So other plugins using those promises should
        use scan_schedule() instead of mainloop_schedule()
        or promise.schedule().
        """
        self.core.call_success("work_queue_add_promise", "scanner", promise)
        return True

    def scan_list_scanners(self, *args, **kwargs):
        with LOCK:
            if len(self.devices_cache) > 0:
                return self.devices_cache

            LOGGER.info("Looking for scan devices ...")
            devs = self.libinsane.list_devices(
                Libinsane.DeviceLocations.ANY
            )
            devs = [
                # (id, human readable name)
                # prefix the IDs with 'libinsane:' so we know it comes from
                # our plugin and not another scan plugin
                (
                    'libinsane:' + dev.get_dev_id(),
                    "{} {}".format(
                        dev.get_dev_vendor(), dev.get_dev_model()
                    ),
                    dev.get_dev_vendor(),
                    dev.get_dev_model(),
                )
                for dev in devs
            ]
            devs.sort(key=lambda s: s[1])
            LOGGER.info("%d devices found: %s", len(devs), devs)
            self.devices_cache = devs
            return devs

    def scan_list_scanners_promise(self):
        """
        Return a promise for listing scanners.
        Must be started with "scan_schedule()", not "promise.schedule()" !
        """
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self.scan_list_scanners
        )

    def scan_get_scanner(self, scanner_dev_id=None):
        if self._last_scanner is not None:
            self._last_scanner.close()
            self._last_scanner = None

        if scanner_dev_id is None:
            scanner_dev_id = self.core.call_success(
                "config_get", "scanner_dev_id"
            )
        if scanner_dev_id is None:
            return None

        if not scanner_dev_id.startswith("libinsane:"):
            return None
        scanner_dev_id = scanner_dev_id[len("libinsane:"):]

        with LOCK:
            LOGGER.info("Accessing scanner '%s' ...", scanner_dev_id)
            scanner = self.libinsane.get_device(scanner_dev_id)
            LOGGER.info("Scanner '%s' opened", scanner.get_name())
            self._last_scanner = Scanner(self.core, scanner)
            return self._last_scanner

    def scan_get_scanner_promise(self, scanner_dev_id=None):
        """
        Return a promise getting a scanner instance.
        Must be started with "scan_schedule()", not "promise.schedule()" !
        """
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self.scan_get_scanner, args=(scanner_dev_id,)
        )

    def scan_promise(self, *args, source_id=None, **kwargs):
        """
        Return a promise that will run a scan.
        Must be started with "scan_schedule()", not "promise.schedule()" !
        """
        scanner_dev_id = self.core.call_success(
            "config_get", "scanner_dev_id"
        )
        if source_id is None:
            source_id = self.core.call_success(
                "config_get", "scanner_source_id"
            )
        if source_id is None:
            raise Exception("No source defined")
        scan_id = next(SCAN_ID_GENERATOR)

        promise = self.core.call_success(
            "scan_get_scanner_promise", scanner_dev_id
        )
        promise = promise.then(
            openpaperwork_core.promise.ThreadedPromise(
                self.core, Scanner.get_source, args=(source_id,)
            )
        )
        promise = promise.then(
            openpaperwork_core.promise.ThreadedPromise(
                self.core, Source.scan, args=(scan_id,), kwargs={
                    'close_on_end': True
                }
            )
        )

        return (scan_id, promise)

    def bug_report_get_attachments(self, out: dict):
        out['scanner'] = {
            'include_by_default': False,
            'date': None,
            'file_type': _("Scanner info."),
            'file_url': _("Select to generate"),
            'file_size': 0,
        }

    def on_bug_report_attachment_selected(self, attachment_id, *args):
        if attachment_id != 'scanner':
            return
        collector = BugReportCollector(self, args)
        collector.run()

    def stats_get(self, out: dict):
        if not LIBINSANE_AVAILABLE:
            out['libinsane_version'] = "not installed"
            return
        out['libinsane_version'] = Libinsane.Api.get_version()



import logging

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_core.promise

CAIRO_AVAILABLE = False
GLIB_AVAILABLE = False

try:
    import cairo
    CAIRO_AVAILABLE = True
except (ImportError, ValueError):
    pass

try:
    from gi.repository import GObject
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    # dummy so chkdeps can still be called
    class GObject(object):
        class SignalFlags(object):
            RUN_LAST = 0

        class GObject(object):
            pass


LOGGER = logging.getLogger(__name__)
BLUR_FACTOR = 8


class BlurRenderer(GObject.GObject):
    __gsignals__ = {
        'getting_size': (GObject.SignalFlags.RUN_LAST, None, ()),
        'size_obtained': (GObject.SignalFlags.RUN_LAST, None, ()),
        'img_obtained': (GObject.SignalFlags.RUN_LAST, None, ()),
    }
    OUTLINE = (0.5, 0.5, 0.5)

    def __init__(
            self, core, work_queue_name, wrapped_renderer):
        super().__init__()
        self.core = core
        self.work_queue_name = work_queue_name
        self.wrapped_renderer = wrapped_renderer
        self.blurry = False

        wrapped_renderer.connect(
            "getting_size", self._on_wrapped_getting_size
        )
        wrapped_renderer.connect(
            "size_obtained", self._on_wrapped_size_obtained
        )
        wrapped_renderer.connect(
            "img_obtained", self._on_wrapped_img_obtained
        )

    def __str__(self):
        return f"BlurRenderer({self.wrapped_renderer})"

    def _on_wrapped_getting_size(self, wrapped_renderer):
        self.emit('getting_size')

    def _on_wrapped_size_obtained(self, wrapped_renderer):
        self.emit('size_obtained')

    def _on_wrapped_img_obtained(self, wrapped_renderer):
        self.emit('img_obtained')

    @property
    def size(self):
        return self.wrapped_renderer.size

    def _set_zoom(self, z):
        self.wrapped_renderer.zoom = z

    def _get_zoom(self):
        return self.wrapped_renderer.zoom

    zoom = property(_get_zoom, _set_zoom)

    def start(self):
        self.wrapped_renderer.start()

    def render(self):
        self.wrapped_renderer.render()

    def hide(self):
        self.wrapped_renderer.hide()

    def close(self):
        self.wrapped_renderer.close()

    def draw(self, cairo_ctx):
        renderer = self.wrapped_renderer

        if not self.blurry:
            return renderer.draw(cairo_ctx)

        zoom = self.zoom / BLUR_FACTOR
        reduced_surface = cairo.ImageSurface(
            cairo.FORMAT_ARGB32,
            int(renderer.size[0] * zoom),
            int(renderer.size[1] * zoom),
        )
        ctx = cairo.Context(reduced_surface)
        ctx.scale(1 / BLUR_FACTOR, 1 / BLUR_FACTOR)
        renderer.draw(ctx)

        cairo_ctx.save()
        try:
            cairo_ctx.scale(BLUR_FACTOR, BLUR_FACTOR)
            cairo_ctx.set_source_surface(reduced_surface)
            cairo_ctx.paint()
        finally:
            cairo_ctx.save()

    def blur(self):
        self.blurry = True
        self.wrapped_renderer.blur()

    def unblur(self):
        self.blurry = False
        self.wrapped_renderer.unblur()


if GLIB_AVAILABLE:
    GObject.type_register(BlurRenderer)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 2000

    def get_interfaces(self):
        return [
            'cairo_url',
            'chkdeps',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'urls',
                'defaults': ['openpaperwork_core.urls'],
            },
        ]

    def chkdeps(self, out: dict):
        if not CAIRO_AVAILABLE:
            out['cairo'].update(openpaperwork_core.deps.CAIRO)
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def cairo_renderer_by_url(
            self, work_queue_name, file_url,
            blur=True, **kwargs):
        if not blur:
            return None

        wrapped_renderer = self.core.call_success(
            "cairo_renderer_by_url",
            work_queue_name, file_url, blur=False, **kwargs
        )
        if wrapped_renderer is None:
            LOGGER.error("Unable to get renderer for '%s' !?", file_url)
            return None

        return BlurRenderer(
            self.core, work_queue_name, wrapped_renderer
        )



import logging
import os
import sqlite3

from . import PluginBase

LOGGER = logging.getLogger(__name__)

# Beware that we use Sqlite, but sqlite python module is not thread-safe
# --> all the calls to sqlite module functions must happen on the main loop,
# even those in the transactions (which are run in a thread)


class Plugin(PluginBase):
    def get_interfaces(self):
        return ['sqlite']

    def get_deps(self):
        return [
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
        ]

    def sqlite_open(self, db_url, *args, **kwargs):
        LOGGER.info("Opening DB %s ...", db_url)
        db_path = self.core.call_success("fs_unsafe", db_url)
        if 'timeout' not in kwargs:
            kwargs['timeout'] = 60
        sql = sqlite3.connect(db_path, *args, **kwargs)
        sql.execute("pragma journal_mode = wal;")
        sql.execute("pragma synchronous = normal;")
        sql.execute("pragma temp_store = memory;")
        if os.name == "posix":
            sql.execute("pragma mmap_size = 30000000000;")
        return sql

    def sqlite_execute(self, cb, *args, **kwargs):
        return self.core.call_one("mainloop_execute", cb, *args, **kwargs)

    def sqlite_schedule(self, cb, *args, **kwargs):
        return self.core.call_one("mainloop_schedule", cb, *args, **kwargs)

    def sqlite_close(self, db, optimize=True):
        if optimize:
            LOGGER.info("Optimizing database ...")
            db.execute("pragma optimize;")
        db.close()
        return True



import openpaperwork_core


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['icon']

    def get_deps(self):
        return [
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
        ]

    def icon_get_pixbuf(self, icon_name, size_px):
        file_name = "{}_{}.png".format(icon_name, size_px)
        return self.core.call_success(
            "gtk_load_pixbuf", "paperwork_gtk.icon.out", file_name
        )



import gettext
import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps

import paperwork_backend.docimport

from . import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.active_doc_id = None
        self.windows = []

    def get_interfaces(self):
        return [
            'chkdeps',
            'doc_open',
            'gtk_doc_import',
            'gtk_window_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'gtk_dialog_single_entry',
                'defaults': ['openpaperwork_gtk.dialogs.single_entry'],
            },
            {
                'interface': 'import',
                'defaults': [
                    'paperwork_backend.docimport.img',
                    'paperwork_backend.docimport.pdf',
                ],
            },
            {
                'interface': 'notifications',
                'defaults': ['paperwork_gtk.notifications.dialog'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def doc_open(self, doc_id, doc_url):
        if self.core.call_success("is_doc", doc_url) is None:
            # New document --> no need to track this doc id.
            # Importer will create a new document in time
            self.active_doc_id = None
            return
        self.active_doc_id = doc_id

    def doc_close(self):
        self.active_doc_id = None

    def _show_no_importer(self, file_uris):
        msg = (
            _("Don't know how to import '%s'. Sorry.") % (file_uris)
        )
        flags = (
            Gtk.DialogFlags.MODAL |
            Gtk.DialogFlags.DESTROY_WITH_PARENT
        )
        dialog = Gtk.MessageDialog(
            transient_for=self.windows[-1],
            flags=flags,
            message_type=Gtk.MessageType.ERROR,
            buttons=Gtk.ButtonsType.OK,
            text=msg
        )
        dialog.connect("response", lambda dialog, response: dialog.destroy())
        dialog.show_all()

    def _request_password(self, importer, file_import):
        self.core.call_success(
            "gtk_show_dialog_single_entry", self,
            _("PDF password"), "",
            importer=importer, file_import=file_import
        )

    def on_dialog_single_entry_reply(
            self, origin, reply, password, *args, **kwargs):
        if origin != self:
            # Not ours
            return
        if not reply:
            # User cancelled
            return
        importer = kwargs['importer']
        file_import = kwargs['file_import']
        self._do_import(importer, file_import, data={'password': password})
        return True

    def _show_result_no_doc(self):
        msg = _("No new document to import found")
        flags = (
            Gtk.DialogFlags.MODAL |
            Gtk.DialogFlags.DESTROY_WITH_PARENT
        )
        dialog = Gtk.MessageDialog(
            transient_for=self.windows[-1],
            flags=flags,
            message_type=Gtk.MessageType.WARNING,
            buttons=Gtk.ButtonsType.OK,
            text=msg
        )
        dialog.connect(
            "response", lambda dialog, response: dialog.destroy()
        )
        dialog.show_all()

    def _show_result_doc(self, doc_id):
        doc_url = self.core.call_success("doc_id_to_url", doc_id)
        assert doc_url is not None

        if self.active_doc_id != doc_id:
            self.core.call_all("doc_open", doc_id, doc_url)

        nb_pages = self.core.call_success("doc_get_nb_pages_by_url", doc_url)
        if nb_pages is None:
            # may happen with freshly-import .doc/.xsl/.docx/etc files
            nb_pages = 0
        if nb_pages <= 0:
            # empty PDF ?
            LOGGER.warning("Document import %s, but no page in it yet ?!", doc_id)
            return
        self.core.call_success(
            "mainloop_schedule", self.core.call_all,
            "doc_goto_page", nb_pages - 1
        )

    def _delete_files(self, file_uris):
        LOGGER.info("Moving imported file(s) to trash ...")
        for file_uri in file_uris:
            LOGGER.info("Moving %s to trash ...", file_uri)
            self.core.call_success("fs_unlink", file_uri)
        notification = self.core.call_success(
            "get_notification_builder",
            gettext.npgettext(
                "import dialog",
                "Imported file deleted",
                "Imported files deleted",
                len(file_uris)
            ),
        )
        if notification is None:
            return
        notification.set_icon("edit-delete").show()

    def _show_result_notification(self, file_import):
        msg = _("Imported:\n")
        for (k, v) in file_import.stats.items():
            msg += ("- {}: {}\n".format(k, v))
        msg = msg.strip()

        notification = self.core.call_success(
            "get_notification_builder", _("Import successful"),
            need_actions=True
        )
        if notification is not None:
            notification.set_message(
                msg
            ).set_icon(
                "document-new"
            ).add_action(
                "delete", _("Delete imported files"),
                self._delete_files, file_import.imported_files
            ).show()

    def _show_result(self, file_import):
        doc_id = None
        if len(file_import.upd_doc_ids) > 0:
            doc_id = list(file_import.upd_doc_ids)[0]
        if len(file_import.new_doc_ids) > 0:
            doc_id = list(file_import.new_doc_ids)[0]
        if doc_id is None:
            self._show_result_no_doc()
            return
        self._show_result_doc(doc_id)
        self._show_result_notification(file_import)

    def _reload_docs(self, file_import):
        for doc_id in file_import.upd_doc_ids:
            doc_url = self.core.call_success("doc_id_to_url", doc_id)
            self.core.call_all("doc_reload", doc_id, doc_url)

    def _add_to_recent(self, file_uris):
        for file_uri in file_uris:
            if self.core.call_success("fs_isdir", file_uri) is None:
                # If the user imported a file, assume they won't import it
                # twice but they may import again other files from the same
                # directory
                file_uri = self.core.call_success("fs_dirname", file_uri)
            LOGGER.info("Adding %s to recently used files", file_uri)
            Gtk.RecentManager().add_item(file_uri)

    def _log_result(self, file_import):
        LOGGER.info("Import result:")
        LOGGER.info("- Imported files: %s", file_import.imported_files)
        LOGGER.info("- Non-imported files: %s", file_import.ignored_files)
        LOGGER.info("- New documents: %s", file_import.new_doc_ids)
        LOGGER.info("- Updated documents: %s", file_import.upd_doc_ids)
        for (k, v) in file_import.stats.items():
            LOGGER.info("- %s: %s", k, v)

    def _do_import(self, importer, file_import, data=None):
        promise = importer.get_import_promise(data)
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(self._log_result, file_import)
        promise = promise.then(self._show_result, file_import)
        promise = promise.then(self._reload_docs, file_import)
        self.core.call_success("transaction_schedule", promise)

    def gtk_doc_import(self, file_urls):
        LOGGER.info("Importing: %s", file_urls)

        file_import = paperwork_backend.docimport.FileImport(
            file_urls, self.active_doc_id
        )

        importers = []
        self.core.call_all("get_importer", importers, file_import)
        if len(importers) <= 0:
            self._show_no_importer(file_urls)
            return
        # TODO(Jflesch): Should ask the user what must be done if many
        # importers are possible.
        importer = importers[0]

        self._add_to_recent(file_urls)

        required = set()
        for (k, v) in importer.get_required_data().items():
            required.update(v)

        LOGGER.info(
            "Required data to import %s: %s",
            file_urls, required
        )
        if "password" in required:
            self._request_password(importer, file_import)
        else:
            self._do_import(importer, file_import)
        return True
